#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri Feb  4 11:56:18 2022

@author: jarato
"""

from setuptools import find_packages, setup

setup(
name='PyEyeSim',
author='jozsef arato',
version='0.1',
packages=find_packages(),)
